# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations
import numpy as np
from typing import Union, TYPE_CHECKING, List, Dict
import webbrowser
from PySide6 import QtWidgets
from PySide6.QtCore import Qt, QRectF, QRect, QPointF
from PySide6.QtGui import QPen, QCursor, QBrush, QColor
from PySide6.QtWidgets import QMenu, QGraphicsSceneMouseEvent, QGraphicsItem

from VeraGrid.Gui.Diagrams.SchematicWidget.Injections.injections_template_graphics import InjectionTemplateGraphicItem
from VeraGrid.Gui.messages import yes_no_question, warning_msg
from VeraGrid.Gui.gui_functions import add_menu_entry
from VeraGrid.Gui.general_dialogues import ShortCircuitSelector
from VeraGrid.Gui.Diagrams.generic_graphics import (GenericDiagramWidget, ACTIVE, DEACTIVATED,
                                                    FONT_SCALE, TRANSPARENT)
from VeraGrid.Gui.Diagrams.SchematicWidget.terminal_item import BarTerminalItem, HandleItem, RoundTerminalItem
from VeraGrid.Gui.Diagrams.SchematicWidget.Injections.load_graphics import LoadGraphicItem, Load
from VeraGrid.Gui.Diagrams.SchematicWidget.Injections.generator_graphics import GeneratorGraphicItem, Generator
from VeraGrid.Gui.Diagrams.SchematicWidget.Injections.static_generator_graphics import (StaticGeneratorGraphicItem,
                                                                                        StaticGenerator)
from VeraGrid.Gui.Diagrams.SchematicWidget.Injections.battery_graphics import (BatteryGraphicItem, Battery)
from VeraGrid.Gui.Diagrams.SchematicWidget.Injections.shunt_graphics import (ShuntGraphicItem, Shunt)
from VeraGrid.Gui.Diagrams.SchematicWidget.Injections.external_grid_graphics import (ExternalGridGraphicItem,
                                                                                     ExternalGrid)
from VeraGrid.Gui.Diagrams.SchematicWidget.Injections.current_injection_graphics import (
    CurrentInjectionGraphicItem,
    CurrentInjection)
from VeraGrid.Gui.Diagrams.SchematicWidget.Injections.controllable_shunt_graphics import (
    ControllableShuntGraphicItem,
    ControllableShunt)
from VeraGridEngine.enumerations import DeviceType, BusGraphicType
from VeraGridEngine.Devices.types import INJECTION_DEVICE_TYPES
from VeraGridEngine.Devices.Substation import Bus
from VeraGridEngine.Devices.Aggregation.short_cirtcuit_event import ShortCircuitEvent

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from VeraGrid.Gui.Diagrams.SchematicWidget.schematic_widget import SchematicWidget

INJECTION_GRAPHICS = Union[
    BatteryGraphicItem,
    ShuntGraphicItem,
    ExternalGridGraphicItem,
    ControllableShuntGraphicItem,
    LoadGraphicItem,
    GeneratorGraphicItem,
    StaticGeneratorGraphicItem,
    CurrentInjectionGraphicItem
]


class ShortCircuitFlags:
    """
    Short circuit flags
    """

    def __init__(self, sc_3p: bool = False, sc_lg: bool = False, sc_ll: bool = False, sc_llg: bool = False):
        self.sc_3p = sc_3p
        self.sc_lg = sc_lg
        self.sc_ll = sc_ll
        self.sc_llg = sc_llg

    def disable_all(self):
        """

        :return:
        """
        self.sc_3p = False
        self.sc_lg = False
        self.sc_ll = False
        self.sc_llg = False


class BusGraphicItem(GenericDiagramWidget, QtWidgets.QGraphicsRectItem):
    """
      Represents a block in the diagram
      Has an x and y and width and height
      width and height can only be adjusted with a tip in the lower right corner.

      - in and output ports
      - parameters
      - description
    """

    def __init__(self,
                 parent=None,
                 index=0,
                 editor: SchematicWidget = None,
                 bus: Bus = None,
                 h: float = 40,
                 w: float = 80,
                 x: float = 0,
                 y: float = 0,
                 draw_labels: bool = True,
                 r: float = 0):
        """

        :param parent:
        :param index:
        :param editor:
        :param bus:
        :param h:
        :param w:
        :param x:
        :param y:
        """
        GenericDiagramWidget.__init__(self, parent=parent, api_object=bus, editor=editor, draw_labels=draw_labels)
        QtWidgets.QGraphicsRectItem.__init__(self, parent)

        # Label:
        self.label = QtWidgets.QGraphicsTextItem(self.api_object.name if self.api_object is not None else "", self)
        self.label.setDefaultTextColor(ACTIVE['text'])
        self.label.setScale(FONT_SCALE)

        # loads, shunts, generators, etc...
        self._child_graphics: List[INJECTION_GRAPHICS] = list()

        # index
        self.index = index

        self.big_marker = None

        self.connectivity_graph = False

        # connection terminals the block
        if self.api_object.graphic_type == BusGraphicType.BusBar:
            self._terminal = BarTerminalItem('s', parent=self, editor=self.editor)
            self.min_w = 180.0
            self.min_h = 40.0
            self.offset = 20

            self.h = h if h >= self.min_h else self.min_h
            self.w = w if w >= self.min_w else self.min_w
            self.r = r

            # square
            self.tile = QtWidgets.QGraphicsRectItem(0, 0, 20, 20, self)
            self.tile.setOpacity(0.7)

            # Create corner for resize:
            self.sizer = HandleItem(self._terminal, callback=self.change_size)
            self.sizer.setPos(self.w, self.h)
            self.sizer.setFlag(self.GraphicsItemFlag.ItemIsMovable)

        elif self.api_object.graphic_type == BusGraphicType.Connectivity:
            self._terminal = RoundTerminalItem('s', parent=self, editor=self.editor, h=20, w=20)  # , h=self.h))
            self.min_w = 50.0
            self.min_h = 50.0
            self.offset = 20

            self.h = h if h >= self.min_h else self.min_h
            self.w = w if w >= self.min_w else self.min_w
            self.r = r

            self._terminal.setPos(self.w / 2 - self._terminal.w / 2, self.h / 2 - self._terminal.h / 2)
            # self._terminal.setPos(self.w / 2 + 10, self.h / 2 + 10)

            # square
            self.tile = QtWidgets.QGraphicsRectItem(0, 0, 5, 5, self)
            self.tile.setOpacity(1.0)

            # Create corner for resize:
            self.sizer = None

            self.connectivity_graph = True

        else:
            self._terminal = BarTerminalItem('s', parent=self, editor=self._editor)
            self.min_w = 180.0
            self.min_h = 40.0
            self.offset = 20

            self.h = h if h >= self.min_h else self.min_h
            self.w = w if w >= self.min_w else self.min_w
            self.r = r

            # square
            self.tile = QtWidgets.QGraphicsRectItem(0, 0, 20, 20, self)
            self.tile.setOpacity(0.7)

            # Create corner for resize:
            self.sizer = HandleItem(self._terminal, callback=self.change_size)
            self.sizer.setPos(self.w, self.h)
            self.sizer.setFlag(self.GraphicsItemFlag.ItemIsMovable)

        self.y0 = self.h + 40  # y position of the shunt element icons

        # Enabled for short circuit
        self.pen_width = 4

        self._terminal.setPen(QPen(TRANSPARENT, self.pen_width, self.style,
                                   Qt.PenCapStyle.RoundCap, Qt.PenJoinStyle.RoundJoin))

        self.set_tile_color(self.color)

        self.setPen(QPen(TRANSPARENT, self.pen_width, self.style))
        self.setBrush(TRANSPARENT)
        self.setFlags(self.GraphicsItemFlag.ItemIsSelectable | self.GraphicsItemFlag.ItemIsMovable)
        self.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))

        # Update size:
        self.change_size(w=self.w)

        self.set_position(x, y)

    @property
    def api_object(self) -> Bus:
        """

        :return:
        """
        return self._api_object

    @property
    def editor(self) -> SchematicWidget:
        """

        :return:
        """
        return self._editor

    @property
    def terminal(self):
        """

        :return:
        """
        return self._terminal

    def get_associated_branch_graphics(self) -> List[GenericDiagramWidget]:
        """
        Get a list of all associated branch graphics
        :return:
        """
        conn: List[GenericDiagramWidget | INJECTION_GRAPHICS] = self._terminal.get_hosted_graphics()

        return conn

    def get_associated_shunt_graphics(self) -> List[INJECTION_GRAPHICS]:
        """
        List of associated shunt graphics
        :return:
        """
        return self._child_graphics

    def get_associated_widgets(self) -> List[GenericDiagramWidget | INJECTION_GRAPHICS]:
        """
        Get a list of all associated graphics
        :return:
        """
        conn: List[GenericDiagramWidget | INJECTION_GRAPHICS] = self.get_associated_branch_graphics()

        for graphics in self._child_graphics:
            conn.append(graphics)
            conn.append(graphics.nexus)

        return conn

    def get_extra_graphics(self) -> List[QGraphicsItem]:
        return self.get_associated_widgets()

    def get_nexus_point(self) -> QPointF:
        """
        Get the connection point for the children nexus line
        (connection points for loads, shunts, generators, etc.)
        :return: QPointF
        """
        if self.connectivity_graph:
            return QPointF(self.x() + self.rect().width() / 2.0,
                           self.y() + self.rect().height() - self._terminal.h)
        else:
            return QPointF(self.x() + self.rect().width() / 2.0,
                           self.y() + self.rect().height() + self._terminal.h / 2.0)

    def recolour_mode(self) -> None:
        """
        Change the colour according to the system theme
        """
        super().recolour_mode()

        self.label.setDefaultTextColor(ACTIVE['text'])
        self.set_tile_color(self.color)

        for e in self._child_graphics:
            if e is not None:
                e.recolour_mode()

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent):
        """
        On mouse move of this object...
        Args:
            event: QGraphicsSceneMouseEvent inherited
        """
        super().mouseMoveEvent(event)
        self._editor.update_diagram_element(device=self.api_object,
                                            x=self.pos().x(),
                                            y=self.pos().y(),
                                            w=self.w,
                                            h=self.h,
                                            r=self.rotation(),
                                            draw_labels=self.draw_labels,
                                            graphic_object=self)

    def add_big_marker(self, color: Union[None, QColor] = QColor(255, 0, 0, 255), tool_tip_text: str = ""):
        """
        Add a big marker to the bus
        :param color: Qt Color ot the marker
        :param tool_tip_text: tool tip text to display
        """
        if color is not None:
            if self.big_marker is None:
                self.big_marker = QtWidgets.QGraphicsEllipseItem(0, 0, 180, 180, parent=self)

            self.big_marker.setBrush(color)
            self.big_marker.setOpacity(0.5)
            self.big_marker.setToolTip(tool_tip_text)

    def delete_big_marker(self) -> None:
        """
        Delete the big marker
        """
        if self.big_marker is not None:
            self._editor._remove_from_scene(self.big_marker)
            self.big_marker = None

    def set_position(self, x: float, y: float) -> None:
        """
        Set the bus x, y position
        :param x: x in pixels
        :param y: y in pixels
        """
        if np.isnan(x):
            x = 0.0
        if np.isnan(y):
            y = 0.0
        self.setPos(QPointF(x, y))

    def set_tile_color(self, brush: QBrush | QColor) -> None:
        """
        Set the color of the bus
        Args:
            brush:  Qt Color
        """
        if self.tile is not None:
            self.tile.setBrush(brush)
        self._terminal.setBrush(brush)

    def merge(self, other_bus_graphic: "BusGraphicItem") -> None:
        """
        Merge another BusGraphicItem into this
        :param other_bus_graphic: BusGraphicItem
        """
        self._child_graphics += other_bus_graphic._child_graphics

    def update(self, rect: Union[QRectF, QRect] = ...):
        """
        Update the object
        :return:
        """
        self.change_size(w=self.w)

    def set_height(self, h: int):
        """
        Set the height of the
        :param h:
        :return:
        """
        self.setRect(0.0, 0.0, self.w, h)
        self.h = h

    def get_terminal_center(self, val: QPointF) -> QPointF:
        """
        Get the center of the terminal
        :param val: position of a branch point
        :return:
        """
        return self._terminal.get_center_pos(val)

    def change_size(self, w: int | float, dummy: float = 0.0):
        """
        Resize block function
        :param w:
        :param dummy:
        :return:
        """
        # Limit the block size to the minimum size:
        self.w = w if w > self.min_w else self.min_w
        self.setRect(0.0, 0.0, self.w, self.min_h)
        y0 = self.offset
        x0 = 0

        # center label:
        self.label.setPos(self.w + 5, -20)

        # lower
        if not self.connectivity_graph:
            self._terminal.setPos(x0, y0)
            self._terminal.setRect(0, 20, self.w, 10)

        # rearrange children
        self.arrange_children()

        # update editor diagram position
        self._editor.update_diagram_element(device=self.api_object,
                                            x=self.pos().x(),
                                            y=self.pos().y(),
                                            w=self.w,
                                            h=int(self.min_h),
                                            r=self.rotation(),
                                            draw_labels=self.draw_labels,
                                            graphic_object=self)

        return self.w, self.min_h

    def arrange_children(self) -> None:
        """
        This function sorts the load and generators icons
        Returns:
            Nothing
        """
        positions = [e.api_object.get_bus_pos(self.api_object) for e in self._child_graphics]
        positions_sorted = np.argsort(positions)

        # y0 = self.h + 40
        n = len(self._child_graphics)
        inc_x = self.w / (n + 1)
        x = inc_x
        for i in positions_sorted:
            elm = self._child_graphics[i]
            elm.setPos(x - elm.w / 2, self.y0)
            x += inc_x

        # Arrange line positions
        self._terminal.process_callbacks(self.pos() + self._terminal.pos())

    def create_children_widgets(self, injections_by_tpe: Dict[DeviceType, List[INJECTION_DEVICE_TYPES]]):
        """
        Create the icons of the elements that are attached to the API bus object
        Returns: Nothing
        """

        for tpe, dev_list in injections_by_tpe.items():

            if tpe == DeviceType.LoadDevice:
                for elm in dev_list:
                    self.add_load(elm)

            elif tpe == DeviceType.StaticGeneratorDevice:
                for elm in dev_list:
                    self.add_static_generator(elm)

            elif tpe == DeviceType.GeneratorDevice:
                for elm in dev_list:
                    self.add_generator(elm)

            elif tpe == DeviceType.ShuntDevice:
                for elm in dev_list:
                    self.add_shunt(elm)

            elif tpe == DeviceType.BatteryDevice:
                for elm in dev_list:
                    self.add_battery(elm)

            elif tpe == DeviceType.ExternalGridDevice:
                for elm in dev_list:
                    self.add_external_grid(elm)

            elif tpe == DeviceType.CurrentInjectionDevice:
                for elm in dev_list:
                    self.add_current_injection(elm)

            elif tpe == DeviceType.ControllableShuntDevice:
                for elm in dev_list:
                    self.add_controllable_shunt(elm)

            else:
                raise Exception("Unknown device type:" + str(tpe))

        self.arrange_children()

    def contextMenuEvent(self, event: QtWidgets.QGraphicsSceneContextMenuEvent):
        """
        Display context menu
        @param event:
        @return:
        """
        menu = QMenu()
        menu.addSection("Bus")

        add_menu_entry(menu=menu,
                       text="Active",
                       icon_path="",
                       function_ptr=self.enable_disable_toggle,
                       checkeable=True,
                       checked_value=self.api_object.active)

        add_menu_entry(menu=menu,
                       text="Draw labels",
                       icon_path="",
                       function_ptr=self.enable_disable_label_drawing,
                       checkeable=True,
                       checked_value=self.draw_labels)

        add_menu_entry(menu=menu,
                       text="add Short circuit",
                       icon_path=":/Icons/icons/short_circuit_plus.png",
                       function_ptr=self.add_short_circuit)

        add_menu_entry(menu=menu,
                       text="Is a DC bus",
                       icon_path=":/Icons/icons/dc.png",
                       function_ptr=self.enable_disable_dc,
                       checkeable=True,
                       checked_value=self.api_object.is_dc)

        add_menu_entry(menu=menu,
                       text="Plot profiles",
                       icon_path=":/Icons/icons/plot.png",
                       function_ptr=self.plot_profiles)

        add_menu_entry(menu,
                       text='Arrange',
                       icon_path=":/Icons/icons/automatic_layout.png",
                       function_ptr=self.arrange_children)

        # add_menu_entry(menu,
        #                text='Rotate',
        #                # icon_path=":/Icons/icons/automatic_layout.png",
        #                function_ptr=self.rotate)

        add_menu_entry(menu,
                       text='Assign active state to profile',
                       icon_path=":/Icons/icons/assign_to_profile.png",
                       function_ptr=self.assign_status_to_profile)

        add_menu_entry(menu, text='Delete',
                       icon_path=":/Icons/icons/delete_schematic.png",
                       function_ptr=self.delete)

        add_menu_entry(menu, text='Convert to voltage level',
                       icon_path=":/Icons/icons/voltage_level.png",
                       function_ptr=self.convert_to_voltage_level)

        add_menu_entry(menu, text='Convert to impedances circuit',
                       icon_path=":/Icons/icons/voltage_level.png",
                       function_ptr=self.convert_to_connectivity_grid)

        add_menu_entry(menu, text='Expand schematic',
                       icon_path=":/Icons/icons/grid_icon.png",
                       function_ptr=self.expand_diagram_from_bus)

        add_menu_entry(menu, text='Vicinity diagram from here',
                       icon_path=":/Icons/icons/grid_icon.png",
                       function_ptr=self.new_vicinity_diagram_from_here)

        add_menu_entry(menu=menu,
                       text="Open in street view",
                       icon_path=":/Icons/icons/map.png",
                       function_ptr=self.open_street_view)

        menu.addSection("Add")

        # Actions under the "Add" section
        add_menu_entry(menu, text='Load',
                       icon_path=":/Icons/icons/add_load.png",
                       function_ptr=self.add_load)

        add_menu_entry(menu, text='Current injection',
                       icon_path=":/Icons/icons/add_load.png",
                       function_ptr=self.add_current_injection)

        add_menu_entry(menu, text='Shunt',
                       icon_path=":/Icons/icons/add_shunt.png",
                       function_ptr=self.add_shunt)

        add_menu_entry(menu,
                       text='Controllable shunt',
                       icon_path=":/Icons/icons/add_shunt.png",
                       function_ptr=self.add_controllable_shunt)

        add_menu_entry(menu, text='Generator',
                       icon_path=":/Icons/icons/add_gen.png",
                       function_ptr=self.add_generator)

        add_menu_entry(menu, text='Static generator',
                       icon_path=":/Icons/icons/add_stagen.png",
                       function_ptr=self.add_static_generator)

        add_menu_entry(menu, text='Battery',
                       icon_path=":/Icons/icons/add_batt.png",
                       function_ptr=self.add_battery)

        add_menu_entry(menu,
                       text='External grid',
                       icon_path=":/Icons/icons/add_external_grid.png",
                       function_ptr=self.add_external_grid)

        menu.exec_(event.screenPos())

    def assign_status_to_profile(self):
        """
        Assign the snapshot rate to the profile
        """
        self._editor.set_active_status_to_profile(self.api_object)

    def delete(self) -> None:
        """
        Remove this element
        @return:
        """
        deleted, delete_from_db_final = self.editor.delete_with_dialogue(selected=[self], delete_from_db=False)
        if deleted:
            self._terminal.clear()

    def delete_child(self, obj: INJECTION_GRAPHICS | InjectionTemplateGraphicItem):
        """
        Delete a child object
        :param obj:
        """
        self._child_graphics.remove(obj)

    def update_color(self):
        """
        Update the colour
        """
        if self.api_object.active:
            self.set_tile_color(QBrush(ACTIVE['color']))
        else:
            self.set_tile_color(QBrush(DEACTIVATED['color']))

    def convert_to_voltage_level(self) -> None:
        """
        Open the voltage level conversion wizard
        """
        self.editor.convert_busbar_to_voltage_level(bus_graphics=self)

    def convert_to_connectivity_grid(self) -> None:
        """
        Transform this bus into a grid of buses to be able to compute the bus currents
        and delete this bus afterward
        """
        self.editor.transform_busbar_to_connectivity_grid(bus_graphics=self)

    def expand_diagram_from_bus(self) -> None:
        """
        Expands the diagram from this bus
        """
        self._editor.expand_diagram_from_bus(root_bus=self.api_object)

    def new_vicinity_diagram_from_here(self):
        """
        Create new vicinity diagram
        :return:
        """
        if self.api_object is not None:
            self.editor.gui.new_bus_branch_diagram_from_bus(root_bus=self.api_object)
        else:
            warning_msg("The api object is none :(")

    def enable_disable_toggle(self):
        """
        Toggle bus element state
        @return:
        """
        if self.api_object is not None:

            # change the bus state (snapshot)
            self.api_object.active = not self.api_object.active

            # change the Branches state (snapshot)
            for host in self._terminal.hosting_connections:
                if host.api_object is not None:
                    host.set_enable(val=self.api_object.active)

            if not self.api_object.active:
                self.clear_label()

            self.update_color()

            if self._editor.circuit.has_time_series:
                ok = yes_no_question('Do you want to update the time series active status accordingly?',
                                     'Update time series active status')

                if ok:
                    # change the bus state (time series)
                    self._editor.set_active_status_to_profile(self.api_object, override_question=True)

                    # change the Branches state (time series)
                    for host in self._terminal.hosting_connections:
                        if host.api_object is not None:
                            self._editor.set_active_status_to_profile(host.api_object, override_question=True)

    def add_short_circuit(self) -> None:
        """
        Enable 3-phase short circuit
        """
        self.w = ShortCircuitSelector()
        self.w.exec()

        if self.w.was_accepted:
            sc = ShortCircuitEvent(
                name=f"{self.api_object.name} {self.w.fault.value}",
                device=self.api_object,
                fault_type=self.w.fault,
                method=self.w.method,
                phases=self.w.phases
            )

            self.editor.circuit.add_short_circuit_event(sc)

    def enable_disable_dc(self):
        """
        Activates or deactivates the bus as a DC bus
        """
        if self._api_object.is_dc:
            self._api_object.is_dc = False
        else:
            self._api_object.is_dc = True

    def rotate(self):
        """

        :return:
        """
        self.r += 90
        self.setRotation(self.r)
        self._editor.update_diagram_element(device=self._api_object,
                                            x=self.pos().x(),
                                            y=self.pos().y(),
                                            w=self.w,
                                            h=self.h,
                                            r=self.r,
                                            draw_labels=self.draw_labels,
                                            graphic_object=self)

    def plot_profiles(self) -> None:
        """
        Plot profiles
        """
        # get the index of this object
        i = self._editor.circuit.get_buses().index(self._api_object)
        self._editor.plot_bus(i, self._api_object)

    def mousePressEvent(self, event: QtWidgets.QGraphicsSceneMouseEvent):
        """
        mouse press: display the editor
        :param event: QGraphicsSceneMouseEvent
        """

        if self._api_object.device_type == DeviceType.BusDevice:
            self._editor.set_editor_model(api_object=self._api_object)

    def mouseDoubleClickEvent(self, event: QtWidgets.QGraphicsSceneMouseEvent):
        """
        Mouse double click
        :param event: event object
        """
        title = self._api_object.name if self._api_object is not None else ""
        msg = ""
        self.label.setHtml(f'<html><head/><body><p><span style=" font-size:10pt;">{title}<br/></span>'
                           f'<span style=" font-size:6pt;">{msg}</span></p></body></html>')

        self.setToolTip(msg)

    def get_terminal(self) -> BarTerminalItem:
        """
        Get the hosting terminal of this bus object
        :return: TerminalItem
        """
        return self._terminal

    def add_object(self, api_obj: Union[None, INJECTION_DEVICE_TYPES] = None):
        """
        Add any recognized object
        :param api_obj: EditableDevice
        """

        if api_obj.device_type == DeviceType.GeneratorDevice:
            self.add_generator(api_obj=api_obj)

        elif api_obj.device_type == DeviceType.LoadDevice:
            self.add_load(api_obj=api_obj)

        elif api_obj.device_type == DeviceType.StaticGeneratorDevice:
            self.add_static_generator(api_obj=api_obj)

        elif api_obj.device_type == DeviceType.ShuntDevice:
            self.add_shunt(api_obj=api_obj)

        elif api_obj.device_type == DeviceType.BatteryDevice:
            self.add_battery(api_obj=api_obj)

        elif api_obj.device_type == DeviceType.ExternalGridDevice:
            self.add_external_grid(api_obj=api_obj)

        elif api_obj.device_type == DeviceType.CurrentInjectionDevice:
            self.add_current_injection(api_obj=api_obj)

        elif api_obj.device_type == DeviceType.ControllableShuntDevice:
            self.add_controllable_shunt(api_obj=api_obj)

        else:
            raise Exception("Cannot add device of type {}".format(api_obj.device_type.value))

    def add_child_graphic(self, elm: INJECTION_DEVICE_TYPES, graphic: INJECTION_GRAPHICS):
        """
        Add a api object and its graphic to this bus graphics domain
        :param elm:
        :param graphic:
        :return:
        """
        self._child_graphics.append(graphic)
        self.arrange_children()
        self.editor.graphics_manager.add_device(elm=elm, graphic=graphic)

    def add_load(self, api_obj: Union[Load, None] = None):
        """
        Add load object to bus
        :param api_obj:
        :return:
        """
        if api_obj is None or type(api_obj) is bool:
            api_obj = self._editor.circuit.add_load(bus=self._api_object)

        _grph = LoadGraphicItem(parent=self, api_obj=api_obj, editor=self._editor)
        self.add_child_graphic(elm=api_obj, graphic=_grph)
        return _grph

    def add_shunt(self, api_obj: Union[Shunt, None] = None):
        """
        Add shunt device
        :param api_obj: If None, a new shunt is created
        """
        if api_obj is None or type(api_obj) is bool:
            api_obj = self._editor.circuit.add_shunt(bus=self._api_object)

        _grph = ShuntGraphicItem(parent=self, api_obj=api_obj, editor=self._editor)
        self.add_child_graphic(elm=api_obj, graphic=_grph)
        return _grph

    def add_generator(self, api_obj: Union[Generator, None] = None):
        """
        Add generator
        :param api_obj: if None, a new generator is created
        """
        if api_obj is None or type(api_obj) is bool:
            api_obj = self._editor.circuit.add_generator(bus=self._api_object)

        _grph = GeneratorGraphicItem(parent=self, api_obj=api_obj, editor=self._editor)
        self.add_child_graphic(elm=api_obj, graphic=_grph)
        return _grph

    def add_static_generator(self, api_obj: Union[StaticGenerator, None] = None):
        """
        Add static generator
        :param api_obj: If none, a new static generator is created
        :return:
        """
        if api_obj is None or type(api_obj) is bool:
            api_obj = self._editor.circuit.add_static_generator(bus=self._api_object)

        _grph = StaticGeneratorGraphicItem(parent=self, api_obj=api_obj, editor=self._editor)
        self.add_child_graphic(elm=api_obj, graphic=_grph)

        return _grph

    def add_battery(self, api_obj: Union[Battery, None] = None):
        """

        :param api_obj:
        :return:
        """
        if api_obj is None or type(api_obj) is bool:
            api_obj = self._editor.circuit.add_battery(bus=self._api_object)

        _grph = BatteryGraphicItem(parent=self, api_obj=api_obj, editor=self._editor)
        self.add_child_graphic(elm=api_obj, graphic=_grph)

        return _grph

    def add_external_grid(self, api_obj: Union[ExternalGrid, None] = None):
        """

        :param api_obj:
        :return:
        """
        if api_obj is None or type(api_obj) is bool:
            api_obj = self._editor.circuit.add_external_grid(bus=self._api_object)

        _grph = ExternalGridGraphicItem(parent=self, api_obj=api_obj, editor=self._editor)
        self.add_child_graphic(elm=api_obj, graphic=_grph)
        return _grph

    def add_current_injection(self, api_obj: Union[CurrentInjection, None] = None):
        """

        :param api_obj:
        :return:
        """
        if api_obj is None or type(api_obj) is bool:
            api_obj = self._editor.circuit.add_current_injection(bus=self._api_object)

        _grph = CurrentInjectionGraphicItem(parent=self, api_obj=api_obj, editor=self._editor)
        self.add_child_graphic(elm=api_obj, graphic=_grph)
        return _grph

    def add_controllable_shunt(self, api_obj: Union[ControllableShunt, None] = None):
        """

        :param api_obj:
        :return:
        """
        if api_obj is None or type(api_obj) is bool:
            api_obj = self._editor.circuit.add_controllable_shunt(bus=self._api_object)

        _grph = ControllableShuntGraphicItem(parent=self, api_obj=api_obj, editor=self._editor)
        self.add_child_graphic(elm=api_obj, graphic=_grph)

        return _grph

    def set_values(self, i: int, Vm: float, Va: float, P: float | None, Q: float | None,
                   tpe: str, format_str="{:10.2f}"):
        """

        :param i:
        :param Vm:
        :param Va:
        :param P:
        :param Q:
        :param tpe:
        :param format_str:
        :return:
        """
        if self.draw_labels:

            msg = f"Bus {i}"
            if tpe is not None:
                msg += f" [{tpe}]"
            msg += "<br>"

            vm = format_str.format(Vm)
            vm_kv = format_str.format(Vm * self._api_object.Vnom)
            va = format_str.format(Va)
            msg += f"V={vm_kv} kV<br>  {vm}&lt;{va}º p.u.<br>"

            if P is not None:
                p = format_str.format(P)
                q = format_str.format(Q)
                msg += f"P={p} MW<br>Q={q} MVAr"

        else:
            msg = ""

        title = self._api_object.name if self._api_object is not None else ""
        self.label.setHtml(f'<html><head/><body><p><span style=" font-size:10pt;">{title}<br/></span>'
                           f'<span style=" font-size:6pt;">{msg}</span></p></body></html>')

        self.setToolTip(msg)

    def set_values_3ph(self, i: int,
                       VmA: float, VmB: float, VmC: float,
                       VaA: float, VaB: float, VaC: float,
                       PA: float, PB: float, PC: float,
                       QA: float, QB: float, QC: float,
                       tpe: str, format_str="{:10.2f}"):
        """
        Set three phase tags
        :param i: Bus index
        :param VmA:
        :param VmB:
        :param VmC:
        :param VaA:
        :param VaB:
        :param VaC:
        :param PA:
        :param PB:
        :param PC:
        :param QA:
        :param QB:
        :param QC:
        :param tpe: Bus type
        :param format_str: number formatting string
        """
        if self.draw_labels:

            msg = f"Bus {i}"
            if tpe is not None:
                msg += f" [{tpe}]"
            msg += "<br>"

            for Vm_i, Va_i, ph in zip([VmA, VmB, VmC], [VaA, VaB, VaC], ["a", "b", "c"]):
                if not (Vm_i == 0.0 and Va_i == 0.0):
                    vm = format_str.format(Vm_i)
                    vm_kv = format_str.format(Vm_i * self._api_object.Vnom)
                    va = format_str.format(Va_i)
                    msg += f"V{ph}={vm_kv} kV / {vm}&lt;{va}º p.u.<br>"

            for P_i, Q_i, ph in zip([PA, PB, PC], [QA, QB, QC], ["a", "b", "c"]):
                if not (P_i == 0.0 and Q_i == 0.0):
                    p = format_str.format(P_i)
                    q = format_str.format(Q_i)
                    msg += f"P{ph}={p} MW<br>Q{ph}={q} MVAr<br>"

        else:
            msg = ""

        title = self._api_object.name if self._api_object is not None else ""
        self.label.setHtml(f'<html><head/><body><p><span style=" font-size:10pt;">{title}<br/></span>'
                           f'<span style=" font-size:6pt;">{msg}</span></p></body></html>')

        self.setToolTip(msg)

    def clear_label(self):
        msg = ""
        title = self._api_object.name if self._api_object is not None else ""
        self.label.setHtml(f'<html><head/><body><p><span style=" font-size:10pt;">{title}<br/></span>'
                           f'<span style=" font-size:6pt;">{msg}</span></p></body></html>')

        self.setToolTip(msg)

    def open_street_view(self):
        """
        Call open street maps
        :return:
        """
        # https://maps.google.com/?q=<lat>,<lng>
        if self._api_object is not None:
            url = f"https://www.google.com/maps/?q={self._api_object.latitude},{self._api_object.longitude}"
            webbrowser.open(url)
        else:
            warning_msg(f"No API object available :(")

    def __str__(self):

        if self._api_object is None:
            return f"Bus graphics {hex(id(self))}"
        else:
            return f"Graphics of {self._api_object.name} [{hex(id(self))}]"

    def __repr__(self):
        return str(self)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations


from PySide6 import QtWidgets
from typing import List
from VeraGrid.Gui.SubstationDesigner.substation_designer_gui import Ui_Dialog
from VeraGrid.Gui.object_model import ObjectsModel
from VeraGrid.Gui.messages import yes_no_question
from VeraGridEngine.Devices.multi_circuit import MultiCircuit
from VeraGridEngine.Devices.Substation.voltage_level_template import VoltageLevelTemplate
from VeraGridEngine.enumerations import DeviceType


class SubstationDesigner(QtWidgets.QDialog):
    """
    SubstationDesigner
    """

    def __init__(self, grid: MultiCircuit,
                 default_voltage: float = 10.0,
                 lat: float = 0.0, lon: float = 0.0,
                 parent=None):
        """

        :param parent:
        """
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_Dialog()
        self.ui.setupUi(self)
        self.setWindowTitle('Substation maker')

        self.grid = grid

        self.default_voltage = default_voltage

        self.ui.latdoubleSpinBox.setValue(lat)
        self.ui.londoubleSpinBox.setValue(lon)

        self._accepted = False

        self.ui.se_name_lineEdit.setText(f"Substation {self.grid.get_substation_number()}")
        self.ui.se_code_lineEdit.setText('')

        obj1 = VoltageLevelTemplate(name="VL", voltage=self.default_voltage,
                                    device_type=DeviceType.VoltageLevelTemplate)

        self.property_list = [obj1.property_list[i] for i in [1, 5, 6, 7, 8, 9]]

        self.mdl = ObjectsModel(objects=[obj1],
                                property_list=self.property_list,
                                time_index=None,
                                parent=self.ui.tableView,
                                editable=True)

        self.ui.tableView.setModel(self.mdl)

        self.ui.addVlButton.clicked.connect(self.add_vl)
        self.ui.deleteVlButton.clicked.connect(self.remove_vl)
        self.ui.createButton.clicked.connect(self.create_se)

    def get_name(self):
        """
        Get the SE name
        :return:
        """
        return self.ui.se_name_lineEdit.text()

    def get_code(self):
        """
        Get the SE code
        :return:
        """
        return self.ui.se_code_lineEdit.text()

    def get_selected_objects(self):
        """
        Get the list of selected objects
        :return:
        """

        if self.mdl is not None:
            sel_idx = self.ui.tableView.selectedIndexes()
            if len(sel_idx) > 0:

                # get the unique rows
                unique = set()
                for idx in sel_idx:
                    unique.add(idx.row())

                return [self.mdl.objects[i] for i in unique]
            else:
                return list()
        else:
            return list()

    def add_vl(self) -> None:
        """
        Add voltage level
        :return:
        """
        obj1 = VoltageLevelTemplate(name="VL", voltage=self.default_voltage,
                                    device_type=DeviceType.VoltageLevelTemplate)
        self.mdl.objects.append(obj1)
        self.mdl.update()

    def remove_vl(self):
        """
        Remove voltage level
        :return:
        """
        sel = self.get_selected_objects()
        cpy_lst = self.mdl.objects.copy()
        for obj in sel:
            cpy_lst.remove(obj)

        self.mdl = ObjectsModel(objects=cpy_lst,
                                property_list=self.property_list,
                                time_index=None,
                                parent=self.ui.tableView,
                                editable=True)

        self.ui.tableView.setModel(self.mdl)

    def get_voltage_levels(self) -> List[VoltageLevelTemplate]:
        """
        Get the list of voltage levels
        :return:
        """
        return self.mdl.objects

    def get_latitude(self):
        return self.ui.latdoubleSpinBox.value()

    def get_longitude(self):
        return self.ui.londoubleSpinBox.value()

    def was_ok(self) -> bool:
        """
        Get if to create substation
        """
        return self._accepted

    def create_se(self):
        """
        Create the thing
        :return:
        """
        if len(self.mdl.objects) > 0:
            self._accepted = True
            self.close()
        else:
            ok = yes_no_question("There are no voltage levels, so no substation will be created, ok?")
            if ok:
                self._accepted = False
                self.close()


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    grid_ = MultiCircuit()
    window = SubstationDesigner(grid=grid_)
    # window.resize(int(1.61 * 700.0), int(600.0))  # golden ratio
    window.show()
    sys.exit(app.exec())



# MIT License
#
# Copyright (c) 2021-2022 Yunosuke Ohsugi
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""The ui to show Qt standard icons."""
from __future__ import annotations

from VeraGrid.ThirdParty.qdarktheme.qtpy.QtCore import Qt
from VeraGrid.ThirdParty.qdarktheme.qtpy.QtWidgets import (
    QGridLayout,
    QScrollArea,
    QStyle,
    QToolButton,
    QVBoxLayout,
    QWidget,
)


class IconsUi:
    """The ui class to show Qt standard icons."""

    def setup_ui(self, win: QWidget) -> None:
        """Set up ui."""
        widget_container = QWidget()
        layout = QGridLayout(widget_container)
        standard_pixmap_names = sorted(
            attr for attr in dir(QStyle.StandardPixmap) if attr.startswith("SP_")
        )
        if len(standard_pixmap_names) == 0:
            standard_pixmap_names = sorted(attr for attr in dir(QStyle) if attr.startswith("SP_"))

        for i, name in enumerate(standard_pixmap_names):
            button = QToolButton()
            button.setText(f":{name}:")
            button.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)

            pixmap = getattr(QStyle.StandardPixmap, name)
            icon = win.style().standardIcon(pixmap)
            button.setIcon(icon)
            layout.addWidget(button, int(i / 4), i % 4)

        scroll_area = QScrollArea()
        scroll_area.setWidget(widget_container)

        v_main_layout = QVBoxLayout(win)
        v_main_layout.addWidget(scroll_area)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.  
# SPDX-License-Identifier: MPL-2.0

from typing import List
import os
import pandas as pd
import json
from VeraGridEngine.Devices.Branches.line import SequenceLineType, UndergroundLineType
from VeraGridEngine.Devices.Branches.transformer import TransformerType
from VeraGridEngine.Devices.Branches.wire import Wire
from VeraGridEngine.Templates.Rms.bus_rms_template import BusRmsTemplate
from VeraGridEngine.Templates.Rms.generator_rms_template import get_generator_rms_template
from VeraGridEngine.Templates.Rms.line_rms_template import get_line_rms_template
from VeraGridEngine.Templates.Rms.load_rms_template import LoadRmsTemplate
from VeraGridEngine.Devices.Dynamic.rms_template import RmsModelTemplate
from VeraGridEngine.IO.veragrid.catalogue import (parse_transformer_types, parse_cable_types, parse_wire_types,
                                                  parse_sequence_line_types)


def get_transformer_catalogue() -> List[TransformerType]:
    """

    :return:
    """
    here = os.path.dirname(os.path.abspath(__file__))
    fname = os.path.join(here, 'data', 'transformers.csv')

    if os.path.exists(fname):
        df = pd.read_csv(fname)

        return parse_transformer_types(df)
    else:
        return list()


def get_cables_catalogue() -> List[UndergroundLineType]:
    """

    :return:
    """
    here = os.path.dirname(os.path.abspath(__file__))
    fname = os.path.join(here, 'data', 'cables.csv')

    if os.path.exists(fname):
        df = pd.read_csv(fname)

        return parse_cable_types(df)
    else:
        return list()


def get_wires_catalogue() -> List[Wire]:
    """

    :return:
    """
    here = os.path.dirname(os.path.abspath(__file__))
    fname = os.path.join(here, 'data', 'wires.csv')

    if os.path.exists(fname):
        df = pd.read_csv(fname)

        return parse_wire_types(df)
    else:
        return list()


def get_sequence_lines_catalogue() -> List[SequenceLineType]:
    """

    :return:
    """
    here = os.path.dirname(os.path.abspath(__file__))
    fname = os.path.join(here, 'data', 'sequence_lines.csv')

    if os.path.exists(fname):
        df = pd.read_csv(fname)

        return parse_sequence_line_types(df)
    else:
        return list()


def get_rms_model_catalogue() -> List[RmsModelTemplate]:
    """
    Here the list of all rms templates must be returned in a list
    :return:
    """
    return [get_generator_rms_template(),
            get_line_rms_template(),
            LoadRmsTemplate()]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.  
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations
from typing import TYPE_CHECKING
from PySide6.QtGui import QPen, QIcon, QPixmap
from PySide6.QtWidgets import QMenu, QGraphicsTextItem
from VeraGridEngine.Devices.Injections.static_generator import StaticGenerator
from VeraGrid.Gui.Diagrams.generic_graphics import ACTIVE, DEACTIVATED, OTHER, Square, Condenser
from VeraGrid.Gui.Diagrams.SchematicWidget.Injections.injections_template_graphics import InjectionTemplateGraphicItem
from VeraGrid.Gui.messages import yes_no_question

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from VeraGrid.Gui.Diagrams.SchematicWidget.schematic_widget import SchematicWidget


class StaticGeneratorGraphicItem(InjectionTemplateGraphicItem):

    def __init__(self, parent, api_obj: StaticGenerator, editor: "SchematicWidget"):
        """

        :param parent:
        :param api_obj:
        """
        InjectionTemplateGraphicItem.__init__(self,
                                              parent=parent,
                                              api_obj=api_obj,
                                              editor=editor,
                                              device_type_name='static_generator',
                                              w=40,
                                              h=40)
        self.set_glyph(glyph=Square(self, h=40, w=40, label_letter="S", update_nexus_fcn=self.update_nexus))

    @property
    def api_object(self) -> StaticGenerator:
        return self._api_object



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import re

from typing import List, Dict, Union

from PySide6 import QtWidgets, QtCore, QtGui

from VeraGridEngine.Devices.Aggregation.rms_event import RmsEvent
from VeraGridEngine.Devices.types import ALL_DEV_TYPES


class RmsEventEditor(QtWidgets.QDialog):
    def __init__(self, devices: List[ALL_DEV_TYPES], event, parent=None):
        super().__init__(parent)
        self.setWindowTitle("RMS Event Editor")
        self.setMinimumWidth(400)

        self.devices = devices

        # Extract parameters for initial device
        if event.device is not None:
            device = event.device
            self.current_device = device
            self.parameters: List = [p for p in device.rms_model.model.event_dict]
            self.parameters_str: List[str] = [p.name for p in self.parameters]
        else:
            self.parameters: List = []
            self.parameters_str: List[str] = []
            self.current_device = None

        # Initial event values
        param = event.parameter if event.parameter is not None else None
        time = event.time if event.time is not None else 0.0
        value = event.value if event.value is not None else 0.0

        # Store current selections
        self.current_parameter = param
        self.current_time = time
        self.current_value = value

        # Devices names
        self.dev_str = [d.name for d in self.devices]

        # ---- Main Layout ----
        layout = QtWidgets.QVBoxLayout(self)

        # --- Device selector ---
        layout.addWidget(QtWidgets.QLabel("Available devices:"))
        self.dev_selector = QtWidgets.QComboBox()
        self.dev_selector.addItems(self.dev_str)
        layout.addWidget(self.dev_selector)

        # Button to update parameters
        self.select_device_btn = QtWidgets.QPushButton("Select Device")
        layout.addWidget(self.select_device_btn)

        # --- Parameter selector ---
        layout.addWidget(QtWidgets.QLabel("Available parameters:"))
        self.param_selector = QtWidgets.QComboBox()
        self.param_selector.addItems(self.parameters_str)
        layout.addWidget(self.param_selector)

        # --- Target time ---
        layout.addWidget(QtWidgets.QLabel("Target time:"))
        self.time_input = QtWidgets.QDoubleSpinBox()
        self.time_input.setRange(-1e6, 1e6)
        self.time_input.setDecimals(6)
        self.time_input.setValue(time)
        layout.addWidget(self.time_input)

        # --- New value ---
        layout.addWidget(QtWidgets.QLabel("New value:"))
        self.value_input = QtWidgets.QDoubleSpinBox()
        self.value_input.setRange(-1e6, 1e6)
        self.value_input.setDecimals(6)
        self.value_input.setValue(value)
        layout.addWidget(self.value_input)

        # --- 🔹 Buttons layout (OK / Cancel) ---
        button_layout = QtWidgets.QHBoxLayout()
        self.ok_button = QtWidgets.QPushButton("OK")
        self.cancel_button = QtWidgets.QPushButton("Cancel")
        button_layout.addStretch(1)
        button_layout.addWidget(self.ok_button)
        button_layout.addWidget(self.cancel_button)
        layout.addLayout(button_layout)

        # --- Connections ---
        self.select_device_btn.clicked.connect(self.on_select_device)
        self.dev_selector.currentIndexChanged.connect(self.on_device_changed)
        self.param_selector.currentIndexChanged.connect(self.on_param_changed)
        self.time_input.valueChanged.connect(self.on_time_changed)
        self.value_input.valueChanged.connect(self.on_value_changed)
        self.ok_button.clicked.connect(self.accept)       # 🔹 Accepts dialog
        self.cancel_button.clicked.connect(self.reject)   # 🔹 Cancels dialog

        # Set initial selections
        if self.current_device is not None:
            self.dev_selector.setCurrentText(self.current_device.name)
        if param is not None:
            self.param_selector.setCurrentText(param.name)

    def on_select_device(self):
        """Update parameters list when device is selected."""
        index = self.dev_selector.currentIndex()
        self.current_device = self.devices[index]
        self.get_dev_parameters(self.current_device)
        # Update parameter combo
        self.param_selector.clear()
        self.param_selector.addItems([p.name for p in self.parameters])
        # Reset current parameter
        if self.parameters:
            self.current_parameter = self.parameters[0]

    def on_device_changed(self, index):
        self.current_device = self.devices[index]

    def on_param_changed(self, index):
        if 0 <= index < len(self.parameters):
            self.current_parameter = self.parameters[index]

    def on_time_changed(self, value):
        self.current_time = value

    def on_value_changed(self, value):
        self.current_value = value

    def get_dev_parameters(self, current_device):
        self.parameters = [p for p in current_device.rms_model.model.event_dict]

    def get_updated_event(self):
        """Return an updated RmsEvent with the edited fields."""
        return RmsEvent(
            device=self.current_device,
            parameter=self.current_parameter,
            time=self.current_time,
            value=self.current_value
        )

def parse_float(s: str) -> float:
    """
    Parse a float from a string accepting both comma and dot as decimal separators.
    Handles cases like:
      - "1.5" -> 1.5
      - "1,5" -> 1.5
      - "1.234,56" -> 1234.56  (common European)
      - "1,234.56" -> 1234.56  (common US)
    Raises ValueError if it cannot be parsed.
    """
    if s is None:
        raise ValueError("Empty string")

    s = s.strip()
    if s == "":
        raise ValueError("Empty string")

    # Remove spaces
    s = s.replace(" ", "")

    # If there are both '.' and ',' assume the last one is the decimal separator:
    # common convention: "1.234,56" (dot thousands, comma decimal) -> remove dots, comma->dot
    # or "1,234.56" (comma thousands, dot decimal) -> remove commas, keep dot.
    if '.' in s and ',' in s:
        # decide which occurs last
        if s.rfind(',') > s.rfind('.'):
            # comma is decimal separator
            s = s.replace('.', '')
            s = s.replace(',', '.')
        else:
            # dot is decimal separator
            s = s.replace(',', '')
    else:
        # only comma or only dot or none: normalize comma -> dot
        s = s.replace(',', '.')

    # Now s should look like a normal Python float literal, maybe with +/-
    if not re.fullmatch(r'[+-]?\d+(\.\d+)?', s):
        raise ValueError(f"Not a valid float: {s}")

    return float(s)


class RmsEventDialogue(QtWidgets.QDialog):
    def __init__(self, parameters_list: List[str], target_device_name: str, parent=None):
        super().__init__(parent)
        self.setWindowTitle("RMS Event Editor")
        self.setMinimumWidth(600)

        self.parameters_list = parameters_list
        self.target_device_name = target_device_name

        # ---- Main Layout ----
        layout = QtWidgets.QVBoxLayout(self)

        # --- Device label ---
        label_device = QtWidgets.QLabel(f"<b>Target device:</b> {target_device_name}")
        layout.addWidget(label_device)

        # --- Events Table ---
        self.table = QtWidgets.QTableWidget(0, 4)  # 4 columns: checkbox + 3 data columns
        self.table.setHorizontalHeaderLabels(["", "Parameter", "Time", "New Value"])
        self.table.horizontalHeader().setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeMode.ResizeToContents)
        self.table.horizontalHeader().setStretchLastSection(True)
        self.table.verticalHeader().setVisible(False)
        self.table.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.NoSelection)
        layout.addWidget(self.table)

        # --- Table control buttons ---
        table_button_layout = QtWidgets.QHBoxLayout()
        self.add_row_btn = QtWidgets.QPushButton("➕ Add New Event")
        self.remove_row_btn = QtWidgets.QPushButton("❌ Remove Selected Rows")
        table_button_layout.addWidget(self.add_row_btn)
        table_button_layout.addWidget(self.remove_row_btn)
        layout.addLayout(table_button_layout)

        # --- Bottom buttons ---
        button_layout = QtWidgets.QHBoxLayout()
        self.ok_button = QtWidgets.QPushButton("✅ Add Events")
        self.cancel_button = QtWidgets.QPushButton("Cancel")
        button_layout.addStretch()
        button_layout.addWidget(self.ok_button)
        button_layout.addWidget(self.cancel_button)
        layout.addLayout(button_layout)

        # --- Connections ---
        self.add_row_btn.clicked.connect(self.add_row)
        self.remove_row_btn.clicked.connect(self.remove_checked_rows)
        self.ok_button.clicked.connect(self.accept_dialog)
        self.cancel_button.clicked.connect(self.reject)

    def add_row(self):
        """Add a new empty editable event row."""
        row = self.table.rowCount()
        self.table.insertRow(row)

        # --- Checkbox column ---
        checkbox = QtWidgets.QTableWidgetItem()
        checkbox.setFlags(QtCore.Qt.ItemFlag.ItemIsUserCheckable | QtCore.Qt.ItemFlag.ItemIsEnabled)
        checkbox.setCheckState(QtCore.Qt.CheckState.Unchecked)
        self.table.setItem(row, 0, checkbox)

        # --- Parameter selector ---
        combo = QtWidgets.QComboBox()
        combo.addItems(self.parameters_list)
        self.table.setCellWidget(row, 1, combo)

        # --- Time and Value as plain QLineEdit (no strict validator) ---
        # We'll validate/parse robustly on accept to allow both ',' and '.'
        time_edit = QtWidgets.QLineEdit()
        time_edit.setPlaceholderText("e.g. 1.5 or 1,5")
        value_edit = QtWidgets.QLineEdit()
        value_edit.setPlaceholderText("e.g. 0.95 or 0,95")

        self.table.setCellWidget(row, 2, time_edit)
        self.table.setCellWidget(row, 3, value_edit)

    def remove_checked_rows(self):
        """Remove all rows with checkbox checked."""
        rows_to_remove = []
        for row in range(self.table.rowCount()):
            item = self.table.item(row, 0)
            if item and item.checkState() == QtCore.Qt.CheckState.Checked:
                rows_to_remove.append(row)

        if not rows_to_remove:
            QtWidgets.QMessageBox.information(
                self,
                "No Rows Selected",
                "Please check at least one row to remove.",
            )
            return

        # Remove from bottom to top to avoid index shifting
        for row in reversed(rows_to_remove):
            self.table.removeRow(row)

    def accept_dialog(self):
        """Validate and collect all event data."""
        if self.table.rowCount() == 0:
            QtWidgets.QMessageBox.warning(
                self,
                "No Events",
                "Please add at least one event before confirming.",
            )
            return

        parameters, target_times, values = [], [], []

        for row in range(self.table.rowCount()):
            combo_widget = self.table.cellWidget(row, 1)
            time_widget = self.table.cellWidget(row, 2)
            value_widget = self.table.cellWidget(row, 3)

            param = combo_widget.currentText().strip() if isinstance(combo_widget, QtWidgets.QComboBox) else ""
            time_text = time_widget.text().strip() if isinstance(time_widget, QtWidgets.QLineEdit) else ""
            value_text = value_widget.text().strip() if isinstance(value_widget, QtWidgets.QLineEdit) else ""

            if not param or not time_text or not value_text:
                QtWidgets.QMessageBox.warning(
                    self,
                    "Incomplete Data",
                    f"Row {row + 1} has missing fields.",
                )
                return

            try:
                t = parse_float(time_text)
                v = parse_float(value_text)
            except ValueError as exc:
                QtWidgets.QMessageBox.warning(
                    self,
                    "Invalid Input",
                    f"Row {row + 1} contains invalid numerical values: {exc}",
                )
                return

            parameters.append(param)
            target_times.append(t)
            values.append(v)

        self.data = {
            "parameters": parameters,
            "target_times": target_times,
            "values": values,
        }

        self.accept()

    def get_data(self) -> Dict[str, Union[List[str], List[float]]]:
        """Return collected data."""
        return getattr(self, "data", {"parameters": [], "target_times": [], "values": []})








# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations
from typing import TYPE_CHECKING
from PySide6 import QtWidgets
from PySide6.QtCore import Qt, QPointF
from PySide6.QtGui import QBrush, QColor
from PySide6.QtWidgets import (QMenu, QGraphicsSceneContextMenuEvent, QGraphicsSceneMouseEvent, QGraphicsEllipseItem)

from VeraGrid.Gui.Diagrams.generic_graphics import GenericDiagramWidget
from VeraGrid.Gui.gui_functions import add_menu_entry

from VeraGridEngine.Devices.Substation.voltage_level import VoltageLevel
from VeraGridEngine.Devices.Substation.bus import Bus

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from VeraGrid.Gui.Diagrams.MapWidget.grid_map_widget import GridMapWidget
    from VeraGrid.Gui.Diagrams.MapWidget.Substation.substation_graphic_item import SubstationGraphicItem


class VoltageLevelGraphicItem(GenericDiagramWidget, QGraphicsEllipseItem):
    """
      Represents a block in the diagram
      Has an x and y and width and height
      width and height can only be adjusted with a tip in the lower right corner.

      - in and output ports
      - parameters
      - description
    """

    def __init__(self,
                 parent: SubstationGraphicItem,
                 editor: GridMapWidget,
                 api_object: VoltageLevel,
                 r: float = 0.2,
                 draw_labels: bool = True):
        """

        :param parent:
        :param editor:
        :param api_object:
        :param r:
        :param draw_labels:
        """
        parent_center = parent.get_center_pos()
        GenericDiagramWidget.__init__(self,
                                      parent=parent,
                                      api_object=api_object,
                                      editor=editor,
                                      draw_labels=draw_labels)
        QGraphicsEllipseItem.__init__(self,
                                      parent_center.x(),
                                      parent_center.y(),
                                      r * api_object.Vnom * 0.01,
                                      r * api_object.Vnom * 0.01,
                                      parent)

        parent.register_voltage_level(vl=self)

        self.editor: GridMapWidget = editor  # to reinforce the type

        self.radius = r * api_object.Vnom * 0.01

        self.setAcceptHoverEvents(True)  # Enable hover events for the item

        # Allow moving the node
        self.setFlag(self.GraphicsItemFlag.ItemIsSelectable | self.GraphicsItemFlag.ItemIsMovable)

        self.setCursor(Qt.CursorShape.PointingHandCursor)

        # Create a pen with reduced line width
        pen = self.pen()
        pen.setWidth(0)  # set the pen width to 0
        self.setPen(pen)

        # Create a pen with reduced line width
        se_color = self.api_object.substation.color if self.api_object.substation is not None else QColor("#3d7d95")
        self.color = QColor(se_color)
        self.color.setAlpha(128)
        self.hoover_color = QColor(se_color)
        self.hoover_color.setAlpha(180)
        self.border_color = QColor(se_color)  # No Alpha

        # self.colorInner = QColor(100, random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))
        # self.colorBorder = QColor(100, random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))

        self.colorInner = QColor(50, 80, 175, 100)  # light blue
        self.colorBorder = QColor(36, 59, 131, 100)  # dark blue

        # Assign color to the node
        self.set_color(inner_color=self.colorInner, border_color=self.colorBorder)
        self.hovered = False
        self.needsUpdate = False
        self.setZValue(0)

    @property
    def api_object(self) -> VoltageLevel:
        return self._api_object

    def center_on_substation(self) -> None:
        """
        Centers the graphic item on the substation
        """
        parent_center = self._parent.get_center_pos()
        xc = parent_center.x() - self.rect().width() / 2
        yc = parent_center.y() - self.rect().height() / 2
        self.setRect(xc, yc, self.rect().width(), self.rect().height())

    def move_to_xy(self, x: float, y: float):
        """

        :param x:
        :param y:
        :return:
        """
        self.setRect(x, y, self.rect().width(), self.rect().height())
        return x, y

    def set_size(self, r: float):
        """

        :param r: radius in pixels
        :return:
        """
        # if r != self.radius:
        rect = self.rect()
        rect.setWidth(r)
        rect.setHeight(r)
        self.radius = r

        # change the width and height while keeping the same center
        r2 = r / 2
        new_x = rect.x() - r2
        new_y = rect.y() - r2

        # Set the new rectangle with the updated dimensions
        self.setRect(new_x, new_y, r, r)

    def update_position_at_the_diagram(self) -> None:
        """

        :return:
        """
        real_position = self.pos()
        center_point = self.getPos()
        lat, long = self.editor.to_lat_lon(x=center_point.x() + real_position.x(),
                                           y=center_point.y() + real_position.y())

        self.editor.update_diagram_element(device=self.api_object,
                                           latitude=lat,
                                           longitude=long,
                                           graphic_object=self)

    def get_copy(self, new_parent: SubstationGraphicItem):
        """
        Returns a copy of the VoltageLevelGraphicItem with a different parent
        :param new_parent: New Substation Graphic item parent
        :return: Copy of the VL Graphic Item with new parent.
        """
        return VoltageLevelGraphicItem(parent=new_parent,
                                       editor=new_parent.editor,
                                       api_object=self.api_object,
                                       r=self.radius / self.api_object.Vnom / 0.01,
                                       draw_labels=self.draw_labels)

    def mouseMoveEvent(self, event: QtWidgets.QGraphicsSceneMouseEvent) -> None:
        """
        Event handler for mouse move events.
        """
        if self.hovered:
            self._parent.mouseMoveEvent(event)

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent):
        """
        Event handler for mouse press events.
        """
        super().mousePressEvent(event)
        self._parent.mousePressEvent(event)
        self.editor.disableMove = True
        self.update_position_at_the_diagram()  # always update

        if self.api_object is not None:
            self.editor.set_editor_model(api_object=self.api_object)

    def mouseReleaseEvent(self, event: QtWidgets.QGraphicsSceneMouseEvent) -> None:
        """
        Event handler for mouse release events.
        """
        super().mouseReleaseEvent(event)
        self._parent.mouseReleaseEvent(event)
        self.editor.disableMove = True

    def hoverEnterEvent(self, event: QtWidgets.QGraphicsSceneHoverEvent) -> None:
        """
        Event handler for when the mouse enters the item.
        """
        # self.set_color(QColor(Qt.GlobalColor.red), QColor(Qt.GlobalColor.red))
        self.set_color(self.hoover_color, self.color)
        self.hovered = True

    def hoverLeaveEvent(self, event: QtWidgets.QGraphicsSceneHoverEvent) -> None:
        """
        Event handler for when the mouse leaves the item.
        """
        self.hovered = False
        self.set_color(self.colorInner, self.colorBorder)

    def contextMenuEvent(self, event: QGraphicsSceneContextMenuEvent):
        """

        :param event:
        """
        menu = QMenu()

        add_menu_entry(menu=menu,
                       text="Add bus",
                       icon_path="",
                       function_ptr=self.add_bus)

        menu.exec_(event.screenPos())

    def set_color(self, inner_color: QColor = None, border_color: QColor = None) -> None:
        """

        :param inner_color:
        :param border_color:
        :return:
        """
        # Example: color assignment
        brush = QBrush(inner_color)
        self.setBrush(brush)

        if border_color is not None:
            pen = self.pen()
            pen.setColor(border_color)
            self.setPen(pen)

    def getPos(self) -> QPointF:
        """

        :return:
        """
        # Get the bounding rectangle of the ellipse item
        bounding_rect = self.boundingRect()

        # Calculate the center point of the bounding rectangle
        center_point = bounding_rect.center()

        return center_point

    def add_bus(self):
        """
        Add bus
        """
        bus = Bus(name=f"Bus {self.api_object.name}",
                  Vnom=self.api_object.Vnom,
                  substation=self._parent.api_object,
                  voltage_level=self.api_object)

        self.editor.circuit.add_bus(obj=bus)

    def set_default_color(self) -> None:
        """

        :return:
        """
        # Example: color assignment
        self.set_color(self.color, self.border_color)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations
from typing import Union, TYPE_CHECKING
from PySide6.QtCore import Qt
from PySide6.QtGui import QPen, QColor
from PySide6.QtWidgets import QMenu
from VeraGrid.Gui.Diagrams.SchematicWidget.terminal_item import BarTerminalItem, RoundTerminalItem
from VeraGrid.Gui.Diagrams.generic_graphics import GenericDiagramWidget, ACTIVE
from VeraGrid.Gui.messages import yes_no_question
from VeraGrid.Gui.gui_functions import add_menu_entry
from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.line_graphics_template import LineGraphicTemplateItem
from VeraGridEngine.Devices.Fluid.fluid_path import FluidPath
from VeraGridEngine.enumerations import DeviceType

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from VeraGrid.Gui.Diagrams.SchematicWidget.schematic_widget import SchematicWidget


class FluidPathGraphicItem(LineGraphicTemplateItem):
    """
    LineGraphicItem
    """

    def __init__(self,
                 from_port: Union[BarTerminalItem, RoundTerminalItem],
                 to_port: Union[BarTerminalItem, RoundTerminalItem, None],
                 editor: SchematicWidget,
                 width=10,
                 api_object: FluidPath = None,
                 arrow_size=15,
                 draw_labels: bool = True):
        """
        
        :param from_port:
        :param to_port:
        :param editor:
        :param width:
        :param api_object:
        :param arrow_size:
        """
        GenericDiagramWidget.__init__(self, parent=None, api_object=api_object, editor=editor, draw_labels=draw_labels)
        LineGraphicTemplateItem.__init__(self,
                                         from_port=from_port,
                                         to_port=to_port,
                                         editor=editor,
                                         width=width,
                                         api_object=api_object,
                                         arrow_size=arrow_size,
                                         arrow_f_pos=0.5,
                                         arrow_t_pos=0.5)

        # self.style = Qt.CustomDashLine
        self.style = ACTIVE['style']
        self.color = QColor(api_object.color) if api_object is not None else ACTIVE['fluid']

        self.set_colour(color=self.color,
                        w=self.width,
                        style=self.style)

    @property
    def api_object(self) -> FluidPath:
        return self._api_object

    def set_api_object_color(self):
        """
        Gather the color from the api object and apply
        :return:
        """
        self.color = QColor(self.api_object.color) if self.api_object is not None else ACTIVE['fluid']
        self.set_colour(color=self.color,
                        w=self.width,
                        style=self.style)

    def set_colour(self, color: QColor, w, style: Qt.PenStyle):
        """
        Set color and style
        :param color: QColor instance
        :param w: width
        :param style: PenStyle instance
        :return:
        """

        pen = QPen(color, w, style, Qt.PenCapStyle.RoundCap, Qt.PenJoinStyle.RoundJoin)
        # pen.setDashPattern([5, 3, 2, 3])

        self.setPen(pen)
        self.arrow_p_from.set_colour(color)
        self.arrow_q_from.set_colour(color)
        self.arrow_p_to.set_colour(color)
        self.arrow_q_to.set_colour(color)

        if self.symbol is not None:
            self.symbol.set_colour(color, w, style)

    def recolour_mode(self) -> None:
        """
        Change the colour according to the system theme
        """
        # super().recolour_mode()
        self.set_colour(self.color, self.width, self.style)

    def mouseDoubleClickEvent(self, event):
        """
        On double-click, edit
        :param event:
        :return:
        """
        if self.api_object is not None:
            if self.api_object.device_type in [DeviceType.Transformer2WDevice, DeviceType.LineDevice]:
                # trigger the editor
                self.edit()
            elif self.api_object.device_type is DeviceType.SwitchDevice:
                # change state
                self.enable_disable_toggle()

    def contextMenuEvent(self, event):
        """
        Show context menu
        @param event:
        @return:
        """
        if self.api_object is not None:
            menu = QMenu()
            menu.addSection("FluidPath")

            menu.addSeparator()

            add_menu_entry(menu=menu,
                           text="Plot profiles",
                           icon_path=":/Icons/icons/plot.png",
                           function_ptr=self.plot_profiles)

            add_menu_entry(menu=menu,
                           text="Delete",
                           icon_path=":/Icons/icons/delete3.png",
                           function_ptr=self.delete)

            menu.addSection('Convert to')
            add_menu_entry(menu=menu,
                           text="Convert to line",
                           icon_path=":/Icons/icons/assign_to_profile.png",
                           function_ptr=self.to_line)

            menu.exec_(event.screenPos())
        else:
            pass

    def plot_profiles(self):
        """
        Plot the time series profiles
        @return:
        """
        # get the index of this object
        # i = self.editor.circuit.get_fluid_paths().index(self.api_object)
        # self.editor.diagramScene.plot_branch(i, self.api_object)
        pass

    def edit(self):
        """
        Open the appropriate editor dialogue
        :return:
        """
        pass

    def to_line(self):
        """
        Convert this object to transformer
        :return:
        """
        ok = yes_no_question('Are you sure that you want to convert this fluid path into a line?',
                             'Convert fluid path')
        if ok:
            self._editor.convert_fluid_path_to_line(element=self.api_object, item_graphic=self)



# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations

import rlcompleter
from typing import Dict, Any, List

from PySide6.QtCore import Qt, QEvent, QStringListModel
from PySide6.QtGui import QTextCursor, QFont
from PySide6.QtWidgets import QPlainTextEdit, QCompleter
from VeraGrid.Gui.python_highlighter import PythonHighlighter
from VeraGrid.Gui.font_config import CONSOLE_TEXT_SIZE


class PythonCodeEditor(QPlainTextEdit):
    """
    QPlainTextEdit with runtime Python autocomplete using rlcompleter.
    Same mechanism as the interactive console.
    """

    def __init__(
            self,
            parent=None,
            vars_dict: Dict[str, Any] | None = None, ):
        """

        :param parent:
        :param vars_dict:
        """
        super().__init__(parent)

        font = QFont("Consolas", CONSOLE_TEXT_SIZE)
        self.setFont(font)

        self.setLineWrapMode(QPlainTextEdit.LineWrapMode.NoWrap)

        # VERY IMPORTANT: Tab must NOT move focus
        self.setTabChangesFocus(False)

        PythonHighlighter(self.document())

        # Normalize namespace exactly like the console
        self._vars_dict = dict() if vars_dict is None else vars_dict

        self._vars_dict["__builtins__"] = __builtins__  # dict or module, both fine]

        self._namespace = self._normalize_namespace(self._vars_dict)

        # rlcompleter backend (same as console)
        self._completer_backend = rlcompleter.Completer(self._namespace)

        # Qt popup completer
        self._qt_completer = QCompleter(self)
        self._qt_completer.setWidget(self)
        self._qt_completer.setCompletionMode(QCompleter.CompletionMode.PopupCompletion)
        self._qt_completer.setCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        self._qt_completer.activated.connect(self._insert_completion)

        self._model = QStringListModel(self)
        self._qt_completer.setModel(self._model)

        # Remember what we replace
        self._last_prefix = ""

    def add_var(self, name: str, val: Any) -> None:
        """
        Add variable to the interpreter
        :param name: name of the variable
        :param val: value or pointer
        """
        self._vars_dict[name] = val
        self._namespace = self._normalize_namespace(self._vars_dict)
        self._completer_backend = rlcompleter.Completer(self._namespace)

    @staticmethod
    def _normalize_namespace(ns: Dict[str, Any]) -> Dict[str, Any]:
        builtins_obj = ns.get("__builtins__", __builtins__)
        builtins_dict = (
            builtins_obj if isinstance(builtins_obj, dict)
            else builtins_obj.__dict__
        )

        out = dict(ns)
        out["__builtins__"] = builtins_dict
        return out

    # ------------------------------------------------------------------
    # THIS is the key part: intercept Tab at event() level
    # ------------------------------------------------------------------

    def event(self, e: QEvent):
        if e.type() == QEvent.Type.KeyPress:

            popup = self._qt_completer.popup()

            # --------------------------------------------------
            # If completion popup is visible → handle selection
            # --------------------------------------------------
            if popup.isVisible():
                if e.key() in (Qt.Key.Key_Tab, Qt.Key.Key_Return, Qt.Key.Key_Enter):
                    # current = self._qt_completer.currentCompletion()
                    # if current:
                    #     self._insert_completion(current)
                    popup = self._qt_completer.popup()
                    index = popup.currentIndex()
                    if index.isValid():
                        completion = index.data()
                        self._insert_completion(completion)
                    popup.hide()
                    return True

                if e.key() == Qt.Key.Key_Escape:
                    popup.hide()
                    return True

                # Let popup handle navigation keys
                if e.key() in (Qt.Key.Key_Up, Qt.Key.Key_Down, Qt.Key.Key_PageUp, Qt.Key.Key_PageDown,):
                    return super().event(e)

            # --------------------------------------------------
            # No popup → trigger completion
            # --------------------------------------------------
            if e.key() == Qt.Key.Key_Tab and e.modifiers() == Qt.KeyboardModifier.NoModifier:
                self._trigger_completion()
                return True

            if e.key() == Qt.Key.Key_Space and e.modifiers() & Qt.KeyboardModifier.ControlModifier:
                self._trigger_completion()
                return True

        return super().event(e)

    # ------------------------------------------------------------------
    # Completion logic (console-style)
    # ------------------------------------------------------------------

    def _trigger_completion(self):
        prefix = self._extract_prefix()
        if not prefix:
            return

        self._last_prefix = prefix

        matches: List[str] = []
        i = 0
        while True:
            m = self._completer_backend.complete(prefix, i)
            if m is None:
                break
            matches.append(m)
            i += 1

        if not matches:
            return

        # Remove duplicates, keep order
        seen = set()
        uniq = []
        for m in matches:
            if m not in seen:
                seen.add(m)
                uniq.append(m)

        if len(uniq) == 1:
            self._replace_prefix(uniq[0])
            return

        self._model.setStringList(uniq)

        rect = self.cursorRect()
        rect.setWidth(
            self._qt_completer.popup().sizeHintForColumn(0)
            + self._qt_completer.popup().verticalScrollBar().sizeHint().width()
        )
        self._qt_completer.complete(rect)

    # ------------------------------------------------------------------
    # Prefix handling (critical – NOT WordUnderCursor)
    # ------------------------------------------------------------------

    def _extract_prefix(self) -> str:
        cursor = self.textCursor()
        pos = cursor.position()
        text = self.toPlainText()

        i = pos - 1
        while i >= 0:
            ch = text[i]
            if ch.isalnum() or ch in "._":
                i -= 1
            else:
                break

        return text[i + 1:pos]

    def _replace_prefix(self, completion: str):
        cursor = self.textCursor()
        cursor.beginEditBlock()

        cursor.movePosition(
            QTextCursor.MoveOperation.Left,
            QTextCursor.MoveMode.KeepAnchor,
            len(self._last_prefix),
        )
        cursor.removeSelectedText()
        cursor.insertText(completion)

        cursor.endEditBlock()
        self.setTextCursor(cursor)

    def _insert_completion(self, text: str):
        self._replace_prefix(text)


if __name__ == "__main__":
    import sys
    import numpy as np
    from PySide6.QtWidgets import QApplication, QMainWindow


    class ConsoleMainWindow(QMainWindow):
        def __init__(self):
            super().__init__()
            self.setWindowTitle("PySide6 Python Console")
            console = PythonCodeEditor(
                vars_dict={
                    "__builtins__": __builtins__,  # dict or module, both fine
                    "app": self,
                    "np": np,
                },
            )
            self.setCentralWidget(console)


    app = QApplication(sys.argv)
    window = ConsoleMainWindow()
    window.resize(800, 600)
    window.show()
    sys.exit(app.exec())
